import sendgrid
import sendgrid.helpers.mail


class SendGrid:

    def __init__(self, api_key):
        self.client = sendgrid.SendGridAPIClient(api_key=api_key)

    def send(self, from_addr, to_addr, subject, message_text):
        message = sendgrid.helpers.mail.Mail(
            from_email=from_addr,
            to_emails=to_addr,
            subject=subject,
            plain_text_content=message_text,
        )
        response = self.client.send(message)
        headers = response.headers
        self.result = {k: headers[k] for k in headers.keys()}
        return None



import shlex
import subprocess
import threading
import time
from logging import info

from .netstat import listen_ports


class PortWatcher:
    """runs a command when a new listening port appears, kills the command when the port disappears"""

    def __init__(self, command):
        info(f"{self} init")
        self.command = command
        self.ports = listen_ports()
        self.port = None
        self.proc = None
        self.watch_thread = None
        self.stop_event = threading.Event()

    def running(self):
        ret = self.stop_event.is_set() is False
        info(f"{self} running returning {ret}")
        return ret

    def watch(self):
        ports = listen_ports()

        if ports == self.ports:
            info(f"{self} no change in ports")
            ret = None
        elif self.port is None:
            info(f"{self} port is none")
            new_port = ports.difference(self.ports)
            if new_port:
                self.port = new_port.pop()
                self.ports = ports
                info(f"{self} new port {self.port} detected")
                command = self.command.replace("{}", str(self.port))
                self.proc = subprocess.Popen(shlex.split(command))
                info(f"{self} started {command} pid={self.proc.pid}")
            ret = None
        elif self.port in ports:
            info(f"{self} {self.port=} still in {ports=}")
            ret = None
        else:
            info(f"{self} stopping: pid={self.proc.pid}")
            self.proc.terminate()
            ret = self.proc.wait()
            info(f"{self} stopped: pid={self.proc.pid} ret={ret}")
            self.stop_event.set()
        info(f"{self} watch returning {ret}")
        return ret

    def start(self):
        info(f"{self} starting")
        self.watch_thread = threading.Thread(target=self.run)
        self.watch_thread.start()
        info(f"{self} started")

    def stop(self):
        info(f"{self} stopping")
        self.stop_event.set()
        if self.watch_thread is not None:
            self.watch_thread.join()
            self.watch_thread = None
        info(f"{self} stopped")

    def __enter__(self):
        info(f"{self} enter")
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        info(f"{self} exit")
        self.stop()

    def run(self):
        info(f"{self} run: called")
        while self.stop_event.is_set() is False:
            ret = self.watch()
            info(f"{self} run: watch returned {ret}")
            if ret:
                break
            else:
                time.sleep(1)
        info(f"{self} run: returning")



import base64
import tempfile
from email.message import EmailMessage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

from .watcher import PortWatcher

# If modifying these SCOPES, delete the file token.json.
SCOPES = ["https://www.googleapis.com/auth/gmail.send"]


class Gmail:

    def __init__(self, credentials, reset_token=False):
        """credentials is a base64 encoded credentials.json file. (see https://console.cloud.google.com)"""

        self.credentials = credentials
        self.token_file = Path.home() / ".testmail-gmail-token.json"
        if reset_token:
            self.token_file.unlink()

        with PortWatcher("ssh -q -N -R {}:localhost:{} beaker"):
            self.creds = self.authenticate_gmail_api()

            # create gmail api client
            self.service = build("gmail", "v1", credentials=self.creds)

    def authenticate_gmail_api(self):
        """Shows basic usage of the Gmail API."""

        creds = None
        # The file token.json stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first time.
        if self.token_file.is_file():
            creds = Credentials.from_authorized_user_file(str(self.token_file), SCOPES)
        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                with tempfile.NamedTemporaryFile() as tfp:
                    credentials = base64.b64decode(self.credentials.encode())
                    tfp.write(credentials)
                    tfp.flush()
                    tfp.seek(0)
                    flow = InstalledAppFlow.from_client_secrets_file(tfp.name, SCOPES)
                creds = flow.run_local_server(port=0)
                # Save the credentials for the next run
            self.token_file.write_text(creds.to_json())

        return creds

    def send(self, from_addr, to_addr, subject, message_text):
        """Create and send an email message."""
        service = self.service
        message = MIMEMultipart()
        message["to"] = to_addr
        message["from"] = from_addr
        message["subject"] = subject
        msg = MIMEText(message_text)
        message.attach(msg)

        # Encode the message in base64
        raw = base64.urlsafe_b64encode(message.as_bytes())
        raw = raw.decode()
        body = {"raw": raw}

        try:
            self.message = None
            self.message = service.users().messages().send(userId="me", body=body).execute()
            return None
        except Exception as error:
            return error

    def old_send(self, from_addr, to_addr, subject, message_text):
        """send a gmail message"""

        message = EmailMessage()

        message.set_content(message_text)

        message["To"] = to_addr
        message["From"] = from_addr
        message["Subject"] = subject

        # encoded message
        encoded_message = base64.urlsafe_b64encode(message.as_bytes()).decode()

        create_message = {"message": {"raw": encoded_message}}
        # pylint: disable=E1101
        send_message = self.service.users().messages().send(userId="me", body=create_message).execute()

        print(f'Message id: {send_message["id"]}')
        return send_message



import threading
import time


class RepeatedTimer:
    def __init__(self, interval, function, *args, **kwargs):
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.thread = None
        self.stop_event = threading.Event()

    def __enter__(self):
        self.thread = threading.Thread(target=self.run)
        self.thread.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_event.set()
        self.thread.join()

    def run(self):
        while not self.stop_event.is_set():
            self.function(*self.args, **self.kwargs)
            time.sleep(self.interval)



"""Console script for rstms_testmail."""

import json
import subprocess
import sys
from pprint import pformat

import click
import click.core

from .counter import Counter
from .exception_handler import ExceptionHandler
from .gmail import Gmail
from .sendgrid_server import SendGrid
from .settings import Settings
from .shell import _shell_completion
from .smtp_server import SMTPServer
from .version import __timestamp__, __version__

header = f"{__name__.split('.')[0]} v{__version__} {__timestamp__}"


def _ehandler(ctx, option, debug):
    ctx.obj = dict(ehandler=ExceptionHandler(debug))
    ctx.obj["debug"] = debug
    return debug


def fail(msg):
    click.echo("Failed: " + msg, err=True)
    sys.exit(-1)


@click.command("testmail")
@click.version_option(message=header)
@click.option("-d", "--debug", is_eager=True, is_flag=True, callback=_ehandler, help="debug mode")
@click.option(
    "--shell-completion",
    is_flag=False,
    flag_value="[auto]",
    callback=_shell_completion,
    help="configure shell completion",
)
@click.option("-q", "--quiet", is_flag=True)
@click.option("-v", "--verbose", is_flag=True)
@click.option("-f", "--from", "from_addr", envvar="TESTMAIL_FROM")
@click.option("-t", "--to", "to_addr", envvar="TESTMAIL_TO")
@click.option("-s", "--subject", default="test{}")
@click.option("-c", "--set-counter", type=int)
@click.option("-s", "--system", envvar="TESTMAIL_SYSTEM")
@click.option("-e", "--exec", "exec_command", envvar="TESTMAIL_EXEC")
@click.option("-k", "--api-key", envvar="TESTMAIL_API_KEY")
@click.option("-u", "--username", envvar="TESTMAIL_USERNAME")
@click.option("-p", "--password", envvar="TESTMAIL_PASSWORD")
@click.option("-P", "--port", default=465, envvar="TESTMAIL_PORT")
@click.option("-m", "--message", help="message to send or - to read from stdin")
@click.option("--profile", "profile_option", default="default", show_envvar=True, envvar="TESTMAIL_PROFILE")
@click.option("--dryrun", is_flag=True)
@click.option("--reset-token", is_flag=True)
@click.option("-W", "--write-profile", help="profile name to write")
@click.option("-D", "--delete", "delete_profile", is_flag=True, help="profile name to delete")
@click.option("-L", "--list", "list_profiles", is_flag=True, help="profile name to write")
@click.argument("profile", required=False)
@click.pass_context
def cli(  # noqa:
    ctx,
    debug,
    shell_completion,
    quiet,
    verbose,
    dryrun,
    to_addr,
    from_addr,
    set_counter,
    subject,
    message,
    system,
    exec_command,
    username,
    password,
    port,
    api_key,
    profile,
    write_profile,
    reset_token,
    delete_profile,
    list_profiles,
    profile_option,
):
    """send a test email"""

    if profile is None:
        profile = profile_option

    if "{}" in subject:
        counter = Counter(label="testmail")
        count = counter.bump(set_counter)
        subject = subject.replace("{}", str(count))

    settings = Settings(
        profile=f"testmail_{profile}",
        from_addr=from_addr,
        to_addr=to_addr,
        system=system,
        exec_command=exec_command,
        username=username,
        password=password,
        port=port,
        api_key=api_key,
        write_profile=f"testmail_{write_profile}",
    )

    if delete_profile:
        settings.delete()
        sys.exit(0)

    profiles = settings.profiles()

    if list_profiles:
        click.echo("\n".join(profiles))
        sys.exit(0)

    if message in profiles:
        message is None

    if debug:
        click.echo(json.dumps({"config": settings.dict()}, indent=2))

    if settings.exec_command is not None:
        message = subprocess.check_output(settings.exec_command, shell=True).decode()

    if message == "-":
        message = sys.stdin.read()
    elif message is None:
        message = subject

    system = settings.system or ""

    if system == "gmail":
        server = Gmail(settings.password, reset_token)
    elif system == "sendgrid":
        server = SendGrid(settings.api_key)
    elif system.startswith("smtp:") or system.startswith("smtps:"):
        server = SMTPServer(settings.system, settings.port, settings.username, settings.password)
    else:
        fail(f"unknown system: {repr(system)}")

    if dryrun:
        click.echo(pformat(settings.dict()))
        sys.exit(0)
    else:
        error = server.send(settings.from_addr, settings.to_addr, subject, message)

    if error:
        fail(pformat(error))

    if not quiet:
        click.echo(f"sent {subject} to {settings.to_addr}")

    if verbose:
        click.echo(pformat(server.result))

    sys.exit(0)


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



import json
from pathlib import Path


class Settings:
    def __init__(self, **kwargs):
        profile = kwargs.pop("profile", "")
        write_profile = kwargs.pop("write_profile", "")

        dir = kwargs.pop("dir", "~")
        if dir in ["~", None]:
            dir = Path.home()
        else:
            dir = Path(dir)
        self.dir = dir

        self.file = dir / f".{profile}_settings"

        old_settings = self.read()

        self._settings = {}

        for k, v in kwargs.items():
            if v is None:
                v = old_settings.get(k, None)
            self._settings[k] = v
            setattr(self, k, v)

        if write_profile:
            self.write(self._settings, dir / f".{write_profile}_settings")

    def delete(self):
        self.file.unlink()
        return

    def profiles(self):
        profiles = []
        label, _, _ = self.file.name.partition("_")
        for item in self.file.parent.iterdir():
            if not item.is_file():
                continue
            prefix, _, tail = item.name.partition("_")
            if prefix != label:
                continue
            name, _, _ = tail.partition("_")
            if name:
                profiles.append(name)
        return profiles

    def read(self):
        if self.file.is_file():
            content = self.file.read_text()
        else:
            content = "{}"
        return json.loads(content)

    def write(self, settings, file):
        file.write_text(json.dumps(settings, indent=2))

    def dict(self):
        return self._settings



from pprint import pprint

from rstms_testmail import netstat


def test_netstat_listen_ports():
    ports = netstat.listen_ports()
    assert isinstance(ports, set)
    assert ports
    pprint(ports)
